#if defined(WIN32)
#  include <windows.h>
#  include <TCHAR.h>
#endif

#define GLH_EXT_SINGLE_FILE
#include <glh/glh_extensions.h>
#include <glh/glh_glut.h>
#include <GL/glext.h>

#include <iostream>
#include <math.h>
#include <time.h>

// Include the all-important NVAPI header, if supported.
#if defined(WIN32)
#  include "nvapi.h"
#endif

// Macro to refer to a buffer offset amount
#define BUFFER_OFFSET(i) ((char *)NULL + (i))

GLfloat lightPosition[4] = { 25.0f, 2.0f, 5.0f, 1.0f };
clock_t time1, time2;
GLfloat frame = 0;
GLuint object = 0;

// Interleaved normal, color, and position data for a simple half-cube.
GLfloat cubeVertexData[] = {
     //NX     NY     NZ      CR     CG     CB      VX     VY     VZ
     0.0f,  0.0f,  1.0f,   0.5f,  0.2f,  1.0f,   0.5f,  0.5f,  0.5f,
     0.0f,  0.0f,  1.0f,   0.5f,  0.2f,  1.0f,  -0.5f,  0.5f,  0.5f,
     0.0f,  0.0f,  1.0f,   0.5f,  0.2f,  1.0f,  -0.5f, -0.5f,  0.5f,
     0.0f,  0.0f,  1.0f,   0.5f,  0.2f,  1.0f,   0.5f, -0.5f,  0.5f,

     0.0f,  1.0f,  0.0f,   0.5f,  0.2f,  1.0f,   0.5f,  0.5f,  0.5f,
     0.0f,  1.0f,  0.0f,   0.5f,  0.2f,  1.0f,   0.5f,  0.5f, -0.5f,
     0.0f,  1.0f,  0.0f,   0.5f,  0.2f,  1.0f,  -0.5f,  0.5f, -0.5f,
     0.0f,  1.0f,  0.0f,   0.5f,  0.2f,  1.0f,  -0.5f,  0.5f,  0.5f,

    -1.0f,  0.0f,  0.0f,   0.5f,  0.2f,  1.0f,  -0.5f, -0.5f, -0.5f,
    -1.0f,  0.0f,  0.0f,   0.5f,  0.2f,  1.0f,  -0.5f, -0.5f,  0.5f,
    -1.0f,  0.0f,  0.0f,   0.5f,  0.2f,  1.0f,  -0.5f,  0.5f,  0.5f,
    -1.0f,  0.0f,  0.0f,   0.5f,  0.2f,  1.0f,  -0.5f,  0.5f, -0.5f,
};

GLuint updateCount = 0;

static void InitializeVisualState()
{
    // Set the viewing projection
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glFrustum(-0.5f, 0.5f, -0.5f, 0.5f, 1.0f, 3.0f);

    // Position the viewer
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef(0.0f, 0.0f, -2.0f);

    // Position the object
    glRotatef(30.0f, 1.0f, 0.0f, 0.0f);
    glRotatef(30.0f, 0.0f, 1.0f, 0.0f);

    glClearColor(0.2f, 0.2f, 0.2f, 1.0f);

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
}

static void InitializeFunctionPointers()
{
    // The necessary VBO functions are included with OpenGL 1.5
    if (glh_init_extension("GL_VERSION_1_5") == GL_FALSE) {
        // Something has gone terribly wrong.
        assert(!"It appears that OpenGL 1.5 is not supported. Not good.\n");
        exit(1);
    }
}

static void InitializeLighting()
{
    //GLfloat diffuse[4]  = { 0.5f, 0.25f,   0.9f, 0.125f };
    GLfloat specular[4] = { 0.9f, 0.75f, 0.125f, 0.125f };

    //glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, diffuse);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 30);

    glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);

    // This will be updated each frame in the idle loop
    glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);
}

static void InitializeVBO()
{
    // Initialize a simple VBO using a few commands which will allow the Expert
    // Driver to report some information on the VBO setup.
    glGenBuffers(1, &object);

    glBindBuffer(GL_ARRAY_BUFFER, object);
    glBufferData(GL_ARRAY_BUFFER, sizeof(cubeVertexData), cubeVertexData, GL_DYNAMIC_DRAW);

    glEnableClientState(GL_NORMAL_ARRAY);
    glNormalPointer(GL_FLOAT, 9 * sizeof(GLfloat), BUFFER_OFFSET(0));

    glEnableClientState(GL_COLOR_ARRAY);
    glColorPointer(3, GL_FLOAT, 9 * sizeof(GLfloat), BUFFER_OFFSET(3 * sizeof(GLfloat)));

    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3, GL_FLOAT, 9 * sizeof(GLfloat), BUFFER_OFFSET(6 * sizeof(GLfloat)));
}

#if defined(WIN32)
static void ExpertDriverFeedback(unsigned int category, unsigned int messageID, const char *message)
{
    // A simple output mechanism for Expert Driver feedback. Expert Driver
    // feedback comes in self-contained, NULL-terminated strings along with
    // a unique ID for both the category and message.

    printf("GLExpert Feedback: ");

    // We could also filter out or log messages based on the category...
    switch(category) {
    case NVAPI_OGLEXPERT_REPORT_ERROR:
        cout << "OpenGL Error - ";
        break;
    case NVAPI_OGLEXPERT_REPORT_SWFALLBACK:
        cout << "Software Fallback - ";
        break;
    case NVAPI_OGLEXPERT_REPORT_PROGRAM:
        cout << "Program - ";
        break;
    case NVAPI_OGLEXPERT_REPORT_VBO:
        cout << "Vertex Buffer Object - ";
        break;
    case NVAPI_OGLEXPERT_REPORT_FBO:
        cout << "Framebuffer Object - ";
        break;
    default:
        cout << "NOT RECOGNIZED - ";
        break;
    }
    printf("Message 0x%08X\n", messageID);

    // The rest of the string is the actual message
    printf(message);
    printf("\n");
}
#endif

static void InitializeOpenGL()
{
    InitializeFunctionPointers();
    InitializeVisualState();

#if defined(WIN32)
    // Enable the OpenGL Expert Driver functionality using NVAPI
    NvAPI_OGL_ExpertModeSet(30,
                            NVAPI_OGLEXPERT_REPORT_ALL,
                            NVAPI_OGLEXPERT_OUTPUT_TO_DEBUGGER |
                            NVAPI_OGLEXPERT_OUTPUT_TO_CALLBACK,
                            ExpertDriverFeedback);
#endif

    InitializeLighting();
    InitializeVBO();
}

static void DisplayHandler()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);
    glDrawArrays(GL_QUADS, 0, 12);

    glutSwapBuffers();
}

static void IdleHandler()
{
    time2 = clock();
    frame += (float)(time2 - time1)/(float)CLOCKS_PER_SEC;
    time1 = time2;

    lightPosition[0] = 25.f * cosf(fmodf(frame, (2 * 3.14159f)));

    glutPostRedisplay();
}

static void KeyHandler(unsigned char k, int x, int y)
{
    if ((k == 27) || (k == 'q')) {
        exit(0);
    } else if (k == 'u') {
        printf(" -- Updating the VBO -- \n\n");

        // As often as desired, update the vertex data, in an effort to
        // generate a few more Expert Driver messages...
        GLfloat *raw = (GLfloat*)glMapBuffer(GL_ARRAY_BUFFER, GL_WRITE_ONLY);

        switch (updateCount % 3) {
        case 0:
            for (int i = 0; i < 12; i++) {
                raw[(i*9)+3] = 0.0f;
            }
            break;
        case 1:
            for (int i = 0; i < 12; i++) {
                raw[(i*9)+3] = 0.5f;
            }
            break;
        case 2:
            for (int i = 0; i < 12; i++) {
                raw[(i*9)+3] = 1.0f;
            }
            break;
        }
        updateCount++;

        glUnmapBuffer(GL_ARRAY_BUFFER);

        printf("\n");
    }

    glutPostRedisplay();
}

static void ResizeHandler(int w, int h)
{
    if (h == 0) h = 1;
    if (w == 0) w = 1;

    glViewport(0, 0, w, h);

    InitializeVisualState();
}

int main(int argc, char **argv)
{
#if defined(WIN32)
    NvAPI_Status status = NvAPI_Initialize();

    if (status != NVAPI_OK) {
        assert(!"NVAPI unable to initialize properly\n");
        exit(1);
    }
#endif

    glutInit(&argc, argv);
    glutInitWindowSize(256, 256);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB);
    glutCreateWindow("GLExpert using NVAPI");

    InitializeOpenGL();

    glutDisplayFunc(DisplayHandler);
    glutIdleFunc(IdleHandler);
    glutKeyboardFunc(KeyHandler);
    glutReshapeFunc(ResizeHandler);

    time1 = clock();

    glutMainLoop();

    return 0;
}
